@@ -24,6 +24,8 @@ from __future__ import print_function
 from __future__ import unicode_literals
 
 import collections
+
+import glob
 import itertools
 import json
 import logging
@@ -33,8 +35,10 @@ import subprocess
 import tempfile
 import weakref
 import re
+from functools import partial
+from typing import List
 
-__version__ = '1.5.5'
+__version__ = '1.6.0'
 
 _LOGGER = logging.getLogger('tftest')
 
@@ -50,6 +54,33 @@ class TerraformTestError(Exception):
   pass
 
 
+_TG_BOOL_ARGS = [
+  "no_auto_init",
+  "no_auto_retry",
+  "source_update",
+  "ignore_dependency_errors",
+  "ignore_dependency_order",
+  "include_external_dependencies",
+  "check",
+  "debug",
+  'non_interactive',
+  'ignore_external_dependencies',
+]
+
+
+_TG_KV_ARGS = [
+  "iam_role",
+  "config",
+  "tfpath",
+  "working_dir",
+  "download_dir",
+  "source",
+  "exclude_dir",
+  "include_dir",
+  "hclfmt_file",
+]
+
+
 def parse_args(init_vars=None, tf_vars=None, targets=None, **kw):
   """Convert method arguments for use in Terraform commands.
 
@@ -64,49 +95,16 @@ def parse_args(init_vars=None, tf_vars=None, targets=None, **kw):
   """
   cmd_args = []
 
-  if kw.get('tg_config'):
-    cmd_args += ['--terragrunt-config', kw['tg_config']]
-  if kw.get('tg_tfpath'):
-    cmd_args += ['--terragrunt-tfpath', kw['tg_tfpath']]
-  if kw.get('tg_no_auto_init') == True:
-    cmd_args.append('--terragrunt-no-auto-init')
-  if kw.get('tg_no_auto_retry') == True:
-    cmd_args.append('--terragrunt-no-auto-retry')
-  if kw.get('tg_non_interactive'):
-    cmd_args.append('--terragrunt-non-interactive')
-  if kw.get('tg_working_dir'):
-    cmd_args += ['--terragrunt-working-dir', kw['tg_working_dir']]
-  if kw.get('tg_download_dir'):
-    cmd_args += ['--terragrunt-download-dir', kw['tg_download_dir']]
-  if kw.get('tg_source'):
-    cmd_args += ['--terragrunt-source', kw['tg_source']]
-  if kw.get('tg_source_update') == True:
-    cmd_args.append('--terragrunt-source-update')
-  if kw.get('tg_iam_role'):
-    cmd_args += ['--terragrunt-iam-role', kw['tg_iam_role']]
-  if kw.get('tg_ignore_dependency_errors') == True:
-    cmd_args.append('--terragrunt-ignore-dependency-errors')
-  if kw.get('tg_ignore_dependency_order') == True:
-    cmd_args.append('--terragrunt-ignore-dependency-order')
-  if kw.get('tg_ignore_external_dependencies'):
-    cmd_args.append('--terragrunt-ignore-external-dependencies')
-  if kw.get('tg_include_external_dependencies') == False:
-    cmd_args.append('--terragrunt-include-external-dependencies')
+  cmd_args += [f'--terragrunt-{arg.replace("_", "-")}'
+               for arg in _TG_BOOL_ARGS if kw.get(f"tg_{arg}")]
+  for arg in _TG_KV_ARGS:
+    if kw.get(f"tg_{arg}"):
+      cmd_args += [f'--terragrunt-{arg.replace("_", "-")}', kw[f"tg_{arg}"]]
   if kw.get('tg_parallelism'):
-    cmd_args.append('terragrunt-parralism={}'.format(kw['tg_parallelism']))
-  if kw.get('tg_exclude_dir'):
-    cmd_args += ['--terragrunt-exclude-dir', kw['tg_exclude_dir']]
-  if kw.get('tg_include_dir'):
-    cmd_args += ['--terragrunt-include-dir', kw['tg_include_dir']]
-  if kw.get('tg_check') == True:
-    cmd_args.append('--terragrunt-check')
-  if kw.get('tg_hclfmt_file'):
-    cmd_args += ['--terragrunt-hclfmt-file', kw['tg_hclfmt_file']]
+    cmd_args.append(f'--terragrunt-parallelism {kw["tg_parallelism"]}')
   if isinstance(kw.get('tg_override_attr'), dict):
     cmd_args += ['--terragrunt-override-attr={}={}'.format(k, v)
                  for k, v in kw.get('tg_override_attr').items()]
-  if kw.get('tg_debug') == True:
-    cmd_args.append('--terragrunt-debug')
 
   if kw.get('auto_approve'):
     cmd_args.append('-auto-approve')
@@ -139,7 +137,6 @@ def parse_args(init_vars=None, tf_vars=None, targets=None, **kw):
     cmd_args += [("-target={}".format(t)) for t in targets]
   if kw.get('tf_var_file'):
     cmd_args.append('-var-file={}'.format(kw['tf_var_file']))
-
   return cmd_args
 
 
@@ -220,7 +217,7 @@ class TerraformPlanOutput(TerraformJSONBase):
         planned_values.get('root_module', {}))
     self.outputs = TerraformValueDict(planned_values.get('outputs', {}))
     self.resource_changes = dict((v['address'], v)
-                                 for v in self._raw['resource_changes'])
+                                 for v in self._raw.get('resource_changes', {}))
     # there might be no variables defined
     self.variables = TerraformValueDict(raw.get('variables', {}))
 
@@ -282,7 +279,7 @@ class TerraformTest(object):
       relative to basedir.
     basedir: optional base directory to use for relative paths, defaults to the
       directory above the one this module lives in.
-    terraform: path to the Terraform command.
+    binary: path to the Terraform command.
     env: a dict with custom environment variables to pass to terraform.
   """
 
@@ -292,69 +289,38 @@ class TerraformTest(object):
     self.binary = binary
     self.tfdir = self._abspath(tfdir)
     self.env = os.environ.copy()
+    self.tg_run_all = False
+    self._plan_formatter = lambda out: TerraformPlanOutput(json.loads(out))
+    self._output_formatter = lambda out: TerraformValueDict(json.loads(out))
     if env is not None:
       self.env.update(env)
 
   @classmethod
-  def _cleanup(cls, tfdir, filenames, binary, deep=True):
+  def _cleanup(cls, tfdir, filenames, deep=True):
     """Remove linked files, .terraform and/or .terragrunt-cache folder at instance deletion."""
     _LOGGER.debug('cleaning up %s %s', tfdir, filenames)
     for filename in filenames:
       path = os.path.join(tfdir, filename)
-      if os.path.islink(path):
-        os.unlink(path)
+      os.unlink(path)
     if not deep:
       return
-
-    if binary == 'terraform':
-      path = os.path.join(tfdir, '.terraform')
-      if os.path.isdir(path):
-        shutil.rmtree(path)
-      path = os.path.join(tfdir, 'terraform.tfstate')
-      if os.path.isfile(path):
-        os.unlink(path)
-    else:
-      path = os.path.join(tfdir, '.terragrunt-cache')
-      if os.path.isdir(path):
-        shutil.rmtree(path)
-
-  def _plan(self, all=False, output=False,  **kw):
-    """
-    Run Terragrunt or Terraform plan and optionally return the plan output.
-
-    Args:
-      all: Runs Terragrunt command on all subfolders if True
-      output: Returns the output of the plan command
-    """
-
-    cmd_args = parse_args(**kw)
-
-    if 'out' not in kw:
-      with tempfile.NamedTemporaryFile() as fp:
-        cmd_args.append('-out={}'.format(fp.name))
-        kw['out'] = fp.name
-
-    if all:
-      result = self.execute_command('run-all', 'plan', *cmd_args).out
-      if not output:
-        return result
-      return self.execute_command('run-all', 'show', '-no-color', '-json', kw['out'])
-    else:
-      result = self.execute_command('plan', *cmd_args).out
-      if not output:
-        return result
-      return self.execute_command('show', '-no-color', '-json', kw['out'])
+    path = os.path.join(tfdir, '.terraform')
+    if os.path.isdir(path):
+      shutil.rmtree(path)
+    path = os.path.join(tfdir, 'terraform.tfstate')
+    if os.path.isfile(path):
+      os.unlink(path)
+    path = os.path.join(tfdir, '**', '.terragrunt-cache*')
+    for tg_dir in glob.glob(path, recursive=True):
+      if os.path.isdir(tg_dir):
+        shutil.rmtree(tg_dir)
 
   def _abspath(self, path):
     """Make relative path absolute from base dir."""
-
-    # print(inspect.getdoc(self.setup))
     return path if path.startswith('/') else os.path.join(self._basedir, path)
 
-  def setup(self, all=False, extra_files=None, plugin_dir=None, init_vars=None,
-            backend=True, cleanup_on_exit=True, tg_non_interactive=False,
-            tg_source_update=False, tg_config=None, tg_working_dir=None,
-            **kw):
+  def setup(self, extra_files=None, plugin_dir=None, init_vars=None,
+            backend=True, cleanup_on_exit=True, **kw):
     """Setup method to use in test fixtures.
 
     This method prepares a new Terraform environment for testing the module
@@ -380,64 +346,31 @@ class TerraformTest(object):
       if os.path.isfile(link_src):
         link_dst = os.path.join(self.tfdir, filename)
         try:
-          os.symlink(link_src, link_dst)
+          if os.name == 'nt':
+            shutil.copy(link_src, link_dst)
+          else:
+            os.symlink(link_src, link_dst)
+          filenames.append(filename)
         except FileExistsError as e:  # pylint:disable=undefined-variable
           _LOGGER.warning(e)
         else:
           _LOGGER.debug('linked %s', link_src)
-          filenames.append(filename)
       else:
         _LOGGER.warning('no such file {}'.format(link_src))
     self._finalizer = weakref.finalize(
-        self, self._cleanup, self.tfdir, filenames, self.binary, deep=cleanup_on_exit)
+        self, self._cleanup, self.tfdir, filenames, deep=cleanup_on_exit)
+    return self.init(plugin_dir=plugin_dir, init_vars=init_vars, backend=backend, **kw)
 
-    if self.binary == 'terragrunt':
-      return self.tg_init(all=all, init_vars=init_vars,
-                          backend=backend, plugin_dir=plugin_dir,
-                          tg_non_interactive=tg_non_interactive, tg_source_update=tg_source_update,
-                          tg_config=tg_config, tg_working_dir=tg_working_dir, **kw)
-
-    return self.tf_init(plugin_dir=plugin_dir, init_vars=init_vars, backend=backend)
-
-  def tf_init(self, input=False, no_color=True, plugin_dir=None,
-              init_vars=None, backend=True):
-    """Run Terraform or Terragrunt init command."""
-    cmd_args = parse_args(input=input, no_color=no_color,
-                          backend=backend, plugin_dir=plugin_dir,
-                          init_vars=init_vars)
+  def init(self, input=False, color=False, force_copy=False, plugin_dir=None,
+           init_vars=None, backend=True, **kw):
+    """Run Terraform init command."""
+    cmd_args = parse_args(input=input, color=color, backend=backend,
+                          force_copy=force_copy, plugin_dir=plugin_dir,
+                          init_vars=init_vars, **kw)
     return self.execute_command('init', *cmd_args).out
 
-  def tg_init(self, all=False, input=False, no_color=True, plugin_dir=None,
-              init_vars=None, backend=True, tg_non_interactive=True,
-              tg_source_update=False, tg_config=None, tg_working_dir=None, **kw):
-    """Run Terragrunt init command."""
-    cmd_args = parse_args(input=input, no_color=no_color,
-                          backend=backend, plugin_dir=plugin_dir,
-                          init_vars=init_vars, tg_non_interactive=tg_non_interactive,
-                          tg_source_update=tg_source_update, tg_config=tg_config,
-                          tg_working_dir=tg_working_dir, **kw)
-    if all:
-      return self.execute_command('run-all', 'init', *cmd_args).out
-    else:
-      return self.execute_command('init', *cmd_args).out
-
-  def validate(self, no_color=True, json=None):
-    """Run Terraform or Terragrunt validate command."""
-    cmd_args = parse_args(no_color=True, json=None)
-    return self.execute_command('validate', *cmd_args).out
-
-  def tg_validate(self, no_color=True, json=None, tg_non_interactive=True,
-                  tg_source_update=False, tg_config=None, tg_working_dir=None, **kw):
-    """Run Terragrunt validate command."""
-    cmd_args = parse_args(no_color=no_color, json=json, tg_non_interactive=tg_non_interactive,
-                          tg_source_update=tg_source_update, tg_config=tg_config, tg_working_dir=tg_working_dir, **kw)
-    if all:
-      return self.execute_command('run-all', 'validate', *cmd_args).out
-    else:
-      return self.execute_command('validate', *cmd_args).out
-
-  def plan(self, input=False, no_color=True, refresh=True, tf_vars=None,
-           targets=None, output=False, tf_var_file=None):
+  def plan(self, input=False, color=False, refresh=True, tf_vars=None,
+           targets=None, output=False, tf_var_file=None, **kw):
     """
     Run Terraform plan command, optionally returning parsed plan output.
 
@@ -451,61 +384,26 @@ class TerraformTest(object):
       output: Determines if output will be returned.
       tf_var_file: Path to terraform variable configuration file relative to `self.tfdir`.
     """
-    result = self._plan('plan', input=input, no_color=no_color, refresh=refresh,
-                        tf_vars=tf_vars, targets=targets, output=output, tf_var_file=tf_var_file)
-
+    cmd_args = parse_args(input=input, color=color,
+                          refresh=refresh, tf_vars=tf_vars,
+                          targets=targets,  tf_var_file=tf_var_file, **kw)
+    if not output:
+      return self.execute_command('plan', *cmd_args).out
+    with tempfile.NamedTemporaryFile() as fp:
+      fp.close()
+    # for tg we need to specify a temp name that is relative for the output to go into each
+    # of the .terragrunt-cache, then plan / show would work, otherwise it overwrites each other!
+    temp_file = fp.name if len(self._tg_ra()) == 0 else os.path.basename(fp.name)
+    cmd_args.append('-out={}'.format(temp_file))
+    self.execute_command('plan', *cmd_args)
+    result = self.execute_command('show', '-no-color', '-json', temp_file)
     try:
-      return TerraformPlanOutput(json.loads(result.out))
+      return self._plan_formatter(result.out)
     except json.JSONDecodeError as e:
       raise TerraformTestError('Error decoding plan output: {}'.format(e))
 
-  def tg_plan(self, all=False, input=False, no_color=True, refresh=True,
-              tf_vars=None, targets=None, output=False, tf_var_file=None,
-              tg_non_interactive=True, tg_source_update=False, tg_config=None,
-              tg_working_dir=None, **kw):
-    """
-    Run Terragrunt plan command, optionally returning parsed plan output.
-
-    Args:
-      input: Ask for input for variables if not directly set.
-      no_color: If specified, output won't contain any color.
-      refresh: Update state prior to checking for differences.
-      tf_vars: Dict of variables in the Terraform configuration.
-      targets: List of resources to target. Operation will be limited to this resource
-        and its dependencies
-      output: Determines if output will be returned.
-      tf_var_file: Path to terraform variable configuration file relative to `self.tfdir`.
-    """
-    result = self._plan(all=all, output=output,
-                        input=input, no_color=no_color,
-                        refresh=refresh, tf_vars=tf_vars,
-                        targets=targets, tf_var_file=tf_var_file,
-                        tg_non_interactive=tg_non_interactive, tg_source_update=tg_source_update,
-                        tg_config=tg_config, tg_working_dir=tg_working_dir, **kw)
-    if not output:
-      return result
-
-    if all:
-      # TODO: Find better way to parse result other than regex
-      plans = re.split('\n(?=\\{"format_version"\\:)', result.out)
-      plan_output = []
-      for plan in plans:
-        try:
-          out = TerraformPlanOutput(json.loads(plan))
-          # TODO: Find a way to distinguish each plan from each other (couldn't find an attr in `out` to use as a key to pair with `out` value)
-          # for now returns list of tftest.TerraformPlanModule objects
-          plan_output.append(out)
-        except json.JSONDecodeError as e:
-          raise TerraformTestError('Error decoding plan output: {}'.format(e))
-      return plan_output
-    else:
-      try:
-        return TerraformPlanOutput(json.loads(result.out))
-      except json.JSONDecodeError as e:
-        raise TerraformTestError('Error decoding plan output: {}'.format(e))
-
-  def apply(self, input=False, no_color=True, auto_approve=True,
-            tf_vars=None, targets=None, tf_var_file=None):
+  def apply(self, input=False, color=False, auto_approve=True,
+            tf_vars=None, targets=None, tf_var_file=None, **kw):
     """
     Run Terraform apply command.
 
@@ -518,85 +416,37 @@ class TerraformTest(object):
         and its dependencies
       tf_var_file: Path to terraform variable configuration file relative to `self.tfdir`.
     """
-    cmd_args = parse_args(input=input, no_color=no_color,
+    cmd_args = parse_args(input=input, color=color,
                           auto_approve=auto_approve, tf_vars=tf_vars,
-                          targets=targets, tf_var_file=tf_var_file)
+                          targets=targets, tf_var_file=tf_var_file, **kw)
     return self.execute_command('apply', *cmd_args).out
 
-  def tg_apply(self, all=False, input=False, no_color=True, auto_approve=True, tf_vars=None,
-               targets=None, tf_var_file=None, tg_non_interactive=True,
-               tg_source_update=False, tg_config=None, tg_working_dir=None, **kw):
-    """
-    Run Terragrunt apply command.
-
-    Args:
-      input: Ask for input for variables if not directly set.
-      no_color: If specified, output won't contain any color.
-      auto_approve: Skip interactive approval of plan before applying.
-      tf_vars: Dict of variables in the Terraform configuration.
-      targets: List of resources to target. Operation will be limited to this resource
-        and its dependencies
-      tf_var_file: Path to terraform variable configuration file relative to `self.tfdir`.
-    """
-    cmd_args = parse_args(input=input, no_color=no_color,
-                          auto_approve=auto_approve, tf_vars=tf_vars,
-                          targets=targets, tf_var_file=tf_var_file,
-                          tg_non_interactive=tg_non_interactive, tg_source_update=tg_source_update,
-                          tg_config=tg_config, tg_working_dir=tg_working_dir, **kw)
-
-    if all:
-      return self.execute_command('run-all', 'apply', *cmd_args).out
-    else:
-      return self.execute_command('apply', *cmd_args).out
-
-  def _output(self, all=False, name=None, **kw):
-    cmd_args = parse_args(**kw)
+  def output(self, name=None, color=False, json_format=True, **kw):
+    """Run Terraform output command."""
+    cmd_args = []
     if name:
       cmd_args.append(name)
-
-    if all:
-      output = self.execute_command('run-all', 'output', *cmd_args).out
-    else:
-      output = self.execute_command('output', *cmd_args).out
-    return output
-
-  def output(self, name=None, no_color=True, json_format=True):
-    """Run Terraform output command."""
-    output = self._output(name=name, no_color=no_color,
-                          json_format=json_format)
+    cmd_args += parse_args(color=color, json_format=json_format, **kw)
+    output = self.execute_command('output', *cmd_args).out
     _LOGGER.debug('output %s', output)
     if json_format:
       try:
-        output = TerraformValueDict(json.loads(output))
+        output = self._output_formatter(output)
       except json.JSONDecodeError as e:
         _LOGGER.warning('error decoding output: {}'.format(e))
     return output
 
-  def tg_output(self, all=False, name=None, no_color=True, json_format=True,
-                tg_non_interactive=True, tg_source_update=False, tg_config=None,
-                tg_working_dir=None, **kw):
-    """Run Terragrunt output command."""
-    output = self._output(all=all, name=name, no_color=no_color, json_format=json_format,
-                          tg_non_interactive=tg_non_interactive, tg_source_update=tg_source_update,
-                          tg_config=tg_config, tg_working_dir=tg_working_dir, **kw)
-
-    # TODO: Figure out how to parse terragrunt run-all output command to return
-    #       dict of {directory: output}
-    _LOGGER.debug('output %s', output)
-    try:
-      output = TerraformValueDict(json.loads(output))
-    except json.JSONDecodeError as e:
-      _LOGGER.warning('error decoding output: {}'.format(e))
-    return output
-
-  def destroy(self, **kw):
+  def destroy(self, color=False, auto_approve=True, tf_vars=None, targets=None, tf_var_file=None,  **kw):
     """Run Terraform destroy command."""
-    cmd_args = parse_args(**kw)
+    cmd_args = parse_args(color=color, auto_approve=auto_approve,
+                          tf_vars=tf_vars, targets=targets,
+                          tf_var_file=tf_var_file,  **kw)
     return self.execute_command('destroy', *cmd_args).out
 
-  def refresh(self, **kw):
+  def refresh(self, color=False, lock=False, tf_vars=None, targets=None,  **kw):
     """Run Terraform refresh command."""
-    cmd_args = parse_args(**kw)
+    cmd_args = parse_args(color=color, lock=lock,
+                          tf_vars=tf_vars, targets=targets,  **kw)
     return self.execute_command('refresh', *cmd_args).out
 
   def state_pull(self):
@@ -611,9 +461,9 @@ class TerraformTest(object):
   def execute_command(self, cmd, *cmd_args):
     """Run arbitrary Terraform command."""
     _LOGGER.debug([cmd, cmd_args])
-    cmdline = [self.binary, cmd]
+    cmdline = [self.binary, *self._tg_ra(), cmd]
     cmdline += cmd_args
-    print(cmdline)
+    _LOGGER.info(cmdline)
     try:
       p = subprocess.Popen(cmdline, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, cwd=self.tfdir, env=self.env)
@@ -629,3 +479,52 @@ class TerraformTest(object):
       _LOGGER.critical(message)
       raise TerraformTestError(message)
     return TerraformCommandOutput(retcode, out, err)
+
+  def _tg_ra(self) -> List[str]:
+    """if run_all return ['run-all'] else [] """
+    return ['run-all'] if self._is_tg() and self.tg_run_all else []
+
+  def _is_tg(self) -> bool:
+    """based on the binary set determines if we are running terragrunt"""
+    return self.binary.endswith('terragrunt')
+
+
+def _parse_run_all_out(output: str, formatter: TerraformJSONBase) -> str:
+  """
+    run-all output a bunch of jsons back to back in one string(no comma),
+    this convert the output to a valid json (put b2b jsons into a list)
+  Args:
+    output: the back to back jsons in a string
+    formatter: output format, could be TerraformValueDict or TerraformPlanOutput
+  Returns:
+    convert the input into a list that is a valid json
+  """
+  dicts = json.loads("[" + re.sub(r"\}\s*\{", "}, {", output) + "]")
+  return [formatter(d) for d in dicts]
+
+
+class TerragruntTest(TerraformTest):
+
+  def __init__(self, tfdir, basedir=None, binary='terragrunt', env=None, tg_run_all=False):
+    """A helper class that could be used for testing terragrunt
+
+    Most operations that apply to :func:`~TerraformTest` also apply to this class.
+    Notice that to use this class for Terragrunt run-all, `tg_run_all` needs to be set to
+    True.  The class would then only be used just for run-all.  If you need individual
+    Terragrunt module testing, create another instance of this helper with
+    tg_run_all=False (default)
+
+    Args:
+      tfdir: the Terraform module directory to test, either an absolute path, or
+             relative to basedir.
+      basedir: optional base directory to use for relative paths, defaults to the
+               directory above the one this module lives in.
+      binary: (Optional) path to terragrunt command.
+      env: a dict with custom environment variables to pass to terraform.
+      tg_run_all: whether the test is for terragrunt run-all, default to False
+    """
+    TerraformTest.__init__(self, tfdir, basedir, binary, env)
+    self.tg_run_all = tg_run_all
+    if self.tg_run_all:
+      self._plan_formatter = partial(_parse_run_all_out, formatter=TerraformPlanOutput)
+      self._output_formatter = partial(_parse_run_all_out, formatter=TerraformValueDict)
